package jsr.daemon.pausable;

import jsr.daemon.*;

/**
 * abstract pausable daemon base class to be subclassed by pausable daemons
 *
 * All singleton accesses are synchronized via this daemon instance if applicable.
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-16
 */
public abstract class PausableDaemon extends GenericDaemon implements Pausable {

    /**
     * reference to the handler for dispatching listener events
     */
    private PausableEventHandler handler = null;

    /**
     * reference to the service provider control interface of this daemon
     */
    private PausableDaemonControl control = null;
    
    /**
     * flag for keeping track of parking state
     */
    private boolean parked = true;
    
    /**
     * flag for keeping track of proper state
     */
    private boolean proper = false;
    
    /**
     * flag for keeping track of pause state
     */
    private boolean paused = false;
    
    
    /** 
     * default constructor for this daemon
     */
    public PausableDaemon() {
        super();
    }

    /**
     * start method for this daemon
     * method is launched via the control interface and has to be subclassed
     * @exception DaemonException thrown by an implementing daemon if starting or restarting fails
     */
    public abstract void start() throws DaemonException;

    /**
     * pause method for this daemon
     * method is launched via the control interface and has to be subclassed
     */
    public abstract void stop();
    
    /**
     * access to the service provider pausable control interface of this pausable daemon
     * method is called by the container for controlling the life state of this daemon
     * @return service provider control interface of this pausable daemon
     */
    protected DaemonControl getDaemonControl() {
        if (null == control) {
           synchronized(this) { 
               if (null == control) {
                  control = new PausableDaemonControl() {

                                /**
                                 * initialization method called in order to start the daemon
                                 * @exception DaemonException thrown by the daemon if initialization fails
                                 */
                                public void init() 
                                    throws DaemonException {
                                    if (!isActive()) {
                                       synchronized(this) {
                                           if (!isActive()) {
                                              proper = true;
                                              PausableDaemon.this.init();
                                              if (proper) {
                                                 parked = false; 
                                                 PausableDaemon.this.start();
                                                 if (parked) setPaused(proper, null);
                                                 else setActive(proper, null); 
                                              }
                                           }
                                       }
                                    }
                                }
    
                                /**
                                 * start method called in order to start or restart the daemon
                                 * @exception DaemonException thrown by the daemon if starting or restarting fails
                                 */
                                public void start() 
                                    throws DaemonException {
                                    if (isPaused()) {
                                       synchronized(this) {
                                           if (isPaused()) {
                                              parked = false;
                                              PausableDaemon.this.start();
                                              setPaused((parked||(!proper)), null);
                                           }
                                       }
                                    }
                                }

                                /**
                                 * pause method called for temporarily stopping the daemon
                                 */
                                public void stop() {
                                   if (isActive()) {
                                      synchronized(this) {
                                          if (isActive()) {     
                                             PausableDaemon.this.stop();
                                             setPaused(proper, null);
                                          }
                                      }
                                   }
                                }

                                /**
                                 * destruction method called in order to stop and shutdown the daemon
                                 */
                                public void destroy() {
                                    if (PausableDaemon.this.super.isActive()) {
                                       synchronized(this) {
                                           if (PausableDaemon.this.super.isActive()) {
                                              if (isActive()) PausableDaemon.this.stop();
                                              if (proper) {
                                                 PausableDaemon.this.destroy();
                                                 setActive(false, null);
                                              }
                                           }
                                       }
                                    }
                                }
                            };
               }
           }
        }
        return control;
    }
    
    /**
     * method for accessing the event handler of this daemon
     * @return event handler of this pausable daemon
     */
    protected DaemonEventHandler getEventHandler() {
        if (null == handler) {
            synchronized(this) {
                if (null == handler) {
                   handler = new PausableEventHandler();
                }
            }
        }
        return handler;
    }

    /**
     * method for indicating the activity state of this daemon
     * used by this daemon for managing its own activity state
     * after calling down initialization is required for getting this daemon active again
     * @param throwable a throwable for informing listeners about an exception cause
     */
    protected void down(Throwable throwable) {
        proper = false;
        super.down(throwable);
    }
    
    /**
     * method for managing the pause state of this daemon
     * @param paused flag indicating the current pause state
     * @param throwable a throwable for indicating a possible exception
     */
    protected final void setPaused(boolean paused, Throwable throwable) {
        if ((paused != this.paused)&&(super.isActive())) {
           PausableEventHandler handler = (PausableEventHandler)getEventHandler();
           synchronized(handler) {
               if ((paused != this.paused)&&(super.isActive())) {
                  this.paused = paused;
                  handler.paused(this, paused, throwable);
               }
           }
        }
    }

    /**
     * method for indicating the pause state of this daemon
     * used by this daemon for managing its own pause state
     * after calling pause a restart is required for getting this daemon unpaused again
     * @param throwable a throwable for indicating a possible exception
     */
    public void pause(Throwable throwable) {
        parked = true;
        setPaused(proper, throwable);
    }

    /**
     * life state indication of this daemon
     * state is managed by the container and/or by this daemon
     * @return true if and only if the implementing daemon is paused
     */
    public boolean isPaused() {
        return (paused&&(super.isActive()));
    }

    /**
     * life state indication of this daemon
     * state is managed by the container and/or by this daemon
     * @return true if and only if the implementing daemon is active
     */
    public boolean isActive() {
        return ((!paused)&&(super.isActive()));
    }

    /**
     * finalizer method for this pausable daemon
     */
    protected void finalize() throws Throwable {
        getDaemonControl().destroy();
        getEventHandler().removeListeners();
        handler = null;
        control = null;
        super.finalize();
    }
}
